@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2015 Enrico M. Crisostomo
+ * Copyright (c) 2014-2016 Enrico M. Crisostomo
  *
  * This program is free software; you can redistribute it and/or modify it under
  * the terms of the GNU General Public License as published by the Free Software
@@ -16,10 +16,10 @@
 /**
  * @file
  * @brief Main `libfswatch` source file.
- * @copyright Copyright (c) 2014-2015 Enrico M. Crisostomo
+ * @copyright Copyright (c) 2014-2016 Enrico M. Crisostomo
  * @license GNU General Public License v. 3.0
  * @author Enrico M. Crisostomo
- * @version 1.8.0
+ * @version 1.10.0
  */
 /**
  * @mainpage
@@ -112,8 +112,8 @@
  * implement thread-safety when using `libfswatch`, therefore, is segregating
  * access to each monitor instance from a different thread.
  *
- * The C API, a layer above the C++ API, has been designed in order to provide
- * the same basic guarantee:
+ * Similarly, the C API has been designed in order to provide the same
+ * guarantees offered by the C++ API:
  *
  *   - Concurrently manipulating different monitoring sessions is thread safe.
  *   - Concurrently manipulating the same monitoring session is _not_ thread
@@ -275,6 +275,33 @@
 /**
  * @page history History
  *
+ * @section v1001 10:0:1
+ *
+ *   - Fix C99 compatibility in cevent.h by not implying enum.
+ *   - Free session memory.
+ *   - Fix segmentation fault when starting monitor.
+ *   - Add fsw_is_running() function to the C API to check that a monitor is
+ *     running.
+ *   - Fix stop sequence in fsw::fsevents_monitor::run() and in
+ *     fsw::fsevents_monitor::on_stop().
+ *
+ * @section v900 9:0:0
+ *
+ *   - Add fsw::monitor_filter::read_from_file() to load filters from a file.
+ *   - Add fsw_stop_monitor() function to stop a running monitor.
+ *   - Change FSW_HANDLE type.
+ *
+ * @section v600 8:0:2
+ *
+ *   - Add a mutex to protect the fsw::monitor::notify_events() method.
+ *   - Substitute C++ header names with C names in C headers.
+ *
+ * @section v600 7:0:1
+ *
+ *   - fsw::monitor::~monitor(): update to invoke fsw::monitor::stop().
+ *   - Close resources in monitor::on_stop() instead of doing it in destructors.
+ *   - Add inactivity callback.
+ *
  * @section v600 6:0:0
  *
  *   - fsw::monitor::stop(): added.
@@ -380,10 +407,6 @@
 
 #include "gettext_defs.h"
 #include <iostream>
-#ifdef HAVE_CXX_MUTEX
-#  include <mutex>
-#  include <atomic>
-#endif
 #include <ctime>
 #include <cstdlib>
 #include <cstring>
@@ -401,7 +424,6 @@ using namespace fsw;
 
 typedef struct FSW_SESSION
 {
-  FSW_HANDLE handle;
   vector<string> paths;
   fsw_monitor_type type;
   fsw::monitor *monitor;
@@ -415,34 +437,9 @@ typedef struct FSW_SESSION
   vector<fsw_event_type_filter> event_type_filters;
   map<string, string> properties;
   void *data;
-#ifdef HAVE_CXX_ATOMIC
-  atomic<bool> running;
-#endif
 } FSW_SESSION;
 
-static bool srand_initialized = false;
 static bool fsw_libfswatch_verbose = false;
-
-#ifdef HAVE_CXX_UNIQUE_PTR
-static fsw_hash_map<FSW_HANDLE, unique_ptr<FSW_SESSION>> sessions;
-#else
-static fsw_hash_map<FSW_HANDLE, FSW_SESSION *> sessions;
-#endif
-
-#ifdef HAVE_CXX_MUTEX
-#  ifdef HAVE_CXX_UNIQUE_PTR
-static fsw_hash_map<FSW_HANDLE, unique_ptr<mutex>> session_mutexes;
-#  else
-static fsw_hash_map<FSW_HANDLE, mutex *> session_mutexes;
-#  endif
-static std::mutex session_mutex;
-
-#  define SESSION_GUARD \
-    std::lock_guard<std::mutex> session_lock(session_mutex);
-#else
-#  define SESSION_GUARD
-#endif
-
 static FSW_THREAD_LOCAL FSW_STATUS last_error;
 
 // Forward declarations.
@@ -510,7 +507,9 @@ void libfsw_cpp_callback_proxy(const std::vector<event>& events,
     if (!cevt->flags_num) cevt->flags = nullptr;
     else
     {
-      cevt->flags = static_cast<fsw_event_flag *> (malloc(sizeof (fsw_event_flag) * cevt->flags_num));
+      cevt->flags =
+        static_cast<fsw_event_flag *> (
+          malloc(sizeof(fsw_event_flag) * cevt->flags_num));
       if (!cevt->flags) throw int(FSW_ERR_MEMORY);
     }
 
@@ -537,43 +536,10 @@ void libfsw_cpp_callback_proxy(const std::vector<event>& events,
 
 FSW_HANDLE fsw_init_session(const fsw_monitor_type type)
 {
-  SESSION_GUARD;
-
-  if (!srand_initialized)
-  {
-    srand(time(nullptr));
-    srand_initialized = true;
-  }
-
-  int handle;
-
-  do
-  {
-    handle = rand();
-  }
-  while (sessions.find(handle) != sessions.end());
-
   FSW_SESSION *session = new FSW_SESSION{};
-
-  session->handle = handle;
   session->type = type;
 
-  // Store the handle and a mutex to guard access to session instances.
-#ifdef HAVE_CXX_UNIQUE_PTR
-  sessions[handle] = unique_ptr<FSW_SESSION>(session);
-#else
-  sessions[handle] = session;
-#endif
-
-#ifdef HAVE_CXX_MUTEX
-#  ifdef HAVE_CXX_UNIQUE_PTR
-  session_mutexes[handle] = unique_ptr<mutex>(new mutex);
-#  else
-  session_mutexes[handle] = new mutex;
-#  endif
-#endif
-
-  return handle;
+  return session;
 }
 
 int create_monitor(const FSW_HANDLE handle, const fsw_monitor_type type)
@@ -593,8 +559,8 @@ int create_monitor(const FSW_HANDLE handle, const fsw_monitor_type type)
       return fsw_set_last_error(int(FSW_ERR_PATHS_NOT_SET));
 
     fsw_callback_context *context_ptr = new fsw_callback_context;
+    context_ptr->handle = session;
     context_ptr->callback = session->callback;
-    context_ptr->handle = session->handle;
     context_ptr->data = session->data;
 
     monitor *current_monitor = monitor_factory::create_monitor(type,
@@ -620,60 +586,35 @@ FSW_STATUS fsw_add_path(const FSW_HANDLE handle, const char *path)
   if (!path)
     return fsw_set_last_error(int(FSW_ERR_INVALID_PATH));
 
-  try
-  {
-    SESSION_GUARD;
-
-    FSW_SESSION *session = get_session(handle);
-    session->paths.push_back(path);
-  }
-  catch (int error)
-  {
-    return fsw_set_last_error(error);
-  }
+  FSW_SESSION *session = get_session(handle);
+  session->paths.push_back(path);
 
   return fsw_set_last_error(FSW_OK);
 }
 
-FSW_STATUS fsw_add_property(const FSW_HANDLE handle, const char * name, const char * value)
+FSW_STATUS fsw_add_property(const FSW_HANDLE handle,
+                            const char *name,
+                            const char *value)
 {
   if (!name || !value)
     return fsw_set_last_error(FSW_ERR_INVALID_PROPERTY);
 
-  try
-  {
-    SESSION_GUARD;
-
-    FSW_SESSION *session = get_session(handle);
-
-    session->properties[name] = value;
-  }
-  catch (int error)
-  {
-    return fsw_set_last_error(error);
-  }
+  FSW_SESSION *session = get_session(handle);
+  session->properties[name] = value;
 
   return fsw_set_last_error(FSW_OK);
 }
 
-FSW_STATUS fsw_set_callback(const FSW_HANDLE handle, const FSW_CEVENT_CALLBACK callback, void * data)
+FSW_STATUS fsw_set_callback(const FSW_HANDLE handle,
+                            const FSW_CEVENT_CALLBACK callback,
+                            void *data)
 {
   if (!callback)
     return fsw_set_last_error(int(FSW_ERR_INVALID_CALLBACK));
 
-  try
-  {
-    SESSION_GUARD;
-
-    FSW_SESSION *session = get_session(handle);
-
-    session->callback = callback;
-    session->data = data;
-  }
-  catch (int error)
-  {
-    return fsw_set_last_error(error);
-  }
+  FSW_SESSION *session = get_session(handle);
+  session->callback = callback;
+  session->data = data;
 
   return fsw_set_last_error(FSW_OK);
 }
@@ -681,18 +622,8 @@ FSW_STATUS fsw_set_callback(const FSW_HANDLE handle, const FSW_CEVENT_CALLBACK c
 FSW_STATUS fsw_set_allow_overflow(const FSW_HANDLE handle,
                                   const bool allow_overflow)
 {
-  try
-  {
-    SESSION_GUARD;
-
-    FSW_SESSION *session = get_session(handle);
-
-    session->allow_overflow = allow_overflow;
-  }
-  catch (int error)
-  {
-    return fsw_set_last_error(error);
-  }
+  FSW_SESSION *session = get_session(handle);
+  session->allow_overflow = allow_overflow;
 
   return fsw_set_last_error(FSW_OK);
 }
@@ -702,54 +633,25 @@ FSW_STATUS fsw_set_latency(const FSW_HANDLE handle, const double latency)
   if (latency < 0)
     return fsw_set_last_error(int(FSW_ERR_INVALID_LATENCY));
 
-  try
-  {
-    SESSION_GUARD;
-
-    FSW_SESSION *session = get_session(handle);
-
-    session->latency = latency;
-  }
-  catch (int error)
-  {
-    return fsw_set_last_error(error);
-  }
+  FSW_SESSION *session = get_session(handle);
+  session->latency = latency;
 
   return fsw_set_last_error(FSW_OK);
 }
 
 FSW_STATUS fsw_set_recursive(const FSW_HANDLE handle, const bool recursive)
 {
-  try
-  {
-    SESSION_GUARD;
-
-    FSW_SESSION *session = get_session(handle);
-
-    session->recursive = recursive;
-  }
-  catch (int error)
-  {
-    return fsw_set_last_error(error);
-  }
+  FSW_SESSION *session = get_session(handle);
+  session->recursive = recursive;
 
   return fsw_set_last_error(FSW_OK);
 }
 
-FSW_STATUS fsw_set_directory_only(const FSW_HANDLE handle, const bool directory_only)
+FSW_STATUS fsw_set_directory_only(const FSW_HANDLE handle,
+                                  const bool directory_only)
 {
-  try
-  {
-    SESSION_GUARD;
-
-    FSW_SESSION *session = get_session(handle);
-
-    session->directory_only = directory_only;
-  }
-  catch (int error)
-  {
-    return fsw_set_last_error(error);
-  }
+  FSW_SESSION *session = get_session(handle);
+  session->directory_only = directory_only;
 
   return fsw_set_last_error(FSW_OK);
 }
@@ -757,18 +659,8 @@ FSW_STATUS fsw_set_directory_only(const FSW_HANDLE handle, const bool directory_
 FSW_STATUS fsw_set_follow_symlinks(const FSW_HANDLE handle,
                                    const bool follow_symlinks)
 {
-  try
-  {
-    SESSION_GUARD;
-
-    FSW_SESSION *session = get_session(handle);
-
-    session->follow_symlinks = follow_symlinks;
-  }
-  catch (int error)
-  {
-    return fsw_set_last_error(error);
-  }
+  FSW_SESSION *session = get_session(handle);
+  session->follow_symlinks = follow_symlinks;
 
   return fsw_set_last_error(FSW_OK);
 }
@@ -776,18 +668,8 @@ FSW_STATUS fsw_set_follow_symlinks(const FSW_HANDLE handle,
 FSW_STATUS fsw_add_event_type_filter(const FSW_HANDLE handle,
                                      const fsw_event_type_filter event_type)
 {
-  try
-  {
-    SESSION_GUARD;
-
-    FSW_SESSION *session = get_session(handle);
-
-    session->event_type_filters.push_back(event_type);
-  }
-  catch (int error)
-  {
-    return fsw_set_last_error(error);
-  }
+  FSW_SESSION *session = get_session(handle);
+  session->event_type_filters.push_back(event_type);
 
   return fsw_set_last_error(FSW_OK);
 }
@@ -795,78 +677,42 @@ FSW_STATUS fsw_add_event_type_filter(const FSW_HANDLE handle,
 FSW_STATUS fsw_add_filter(const FSW_HANDLE handle,
                           const fsw_cmonitor_filter filter)
 {
-  try
-  {
-    SESSION_GUARD;
-
-    FSW_SESSION *session = get_session(handle);
-
-    session->filters.push_back(
-      {filter.text, filter.type, filter.case_sensitive, filter.extended});
-  }
-  catch (int error)
-  {
-    return fsw_set_last_error(error);
-  }
+  FSW_SESSION *session = get_session(handle);
+  session->filters.push_back(
+    {filter.text, filter.type, filter.case_sensitive, filter.extended});
 
   return fsw_set_last_error(FSW_OK);
 }
 
-#ifdef HAVE_CXX_MUTEX
-
-template<typename T>
-class monitor_start_guard
+bool fsw_is_running(const FSW_HANDLE handle)
 {
-  atomic<T>& a;
-  T val;
+  FSW_SESSION *session = get_session(handle);
 
-public:
+  if (!session->monitor)
+    return false;
 
-  monitor_start_guard(atomic<T>& a,
-                      T val,
-                      memory_order sync = memory_order_seq_cst)
-    : a(a), val(val)
-  {
-  }
-
-  ~monitor_start_guard()
-  {
-    a.store(val, memory_order_release);
-  }
-};
-
-#endif
+  return session->monitor->is_running();
+}
 
 FSW_STATUS fsw_start_monitor(const FSW_HANDLE handle)
 {
   try
   {
-#ifdef HAVE_CXX_MUTEX
-    unique_lock<mutex> session_lock(session_mutex, defer_lock);
-    session_lock.lock();
-#endif
-
     FSW_SESSION *session = get_session(handle);
 
-#ifdef HAVE_CXX_MUTEX
-# ifdef HAVE_CXX_ATOMIC
-    if (session->running.load(memory_order_acquire))
-      return fsw_set_last_error(int(FSW_ERR_MONITOR_ALREADY_RUNNING));
-# endif
-
-#  ifdef HAVE_CXX_UNIQUE_PTR
-    unique_ptr<mutex>& sm = session_mutexes.at(handle);
-    lock_guard<mutex> lock_sm(*sm.get());
-#  else
-    mutex * sm = session_mutexes.at(handle);
-    lock_guard<mutex> lock_sm(*sm);
-#  endif
-
-    session_lock.unlock();
-#endif
-
     if (!session->monitor)
-      create_monitor(handle, session->type);
+    {
+      FSW_STATUS ret = create_monitor(handle, session->type);
+
+      if (ret != FSW_OK)
+        return fsw_set_last_error(ret);
+    }
+
+    if (session->monitor == nullptr) // create_monitor returned OK, but monitor were not created
+      return fsw_set_last_error(FSW_ERR_UNKNOWN_MONITOR_TYPE);
+
+    if (session->monitor->is_running())
+      return fsw_set_last_error(int(FSW_ERR_MONITOR_ALREADY_RUNNING));
 
     session->monitor->set_allow_overflow(session->allow_overflow);
     session->monitor->set_filters(session->filters);
@@ -876,13 +722,6 @@ FSW_STATUS fsw_start_monitor(const FSW_HANDLE handle)
     session->monitor->set_recursive(session->recursive);
     session->monitor->set_directory_only(session->directory_only);
 
-#ifdef HAVE_CXX_MUTEX
-# ifdef HAVE_CXX_ATOMIC
-    session->running.store(true, memory_order_release);
-    monitor_start_guard<bool> guard(session->running, false);
-# endif
-#endif
-
     session->monitor->start();
   }
   catch (libfsw_exception& ex)
@@ -897,28 +736,47 @@ FSW_STATUS fsw_start_monitor(const FSW_HANDLE handle)
   return fsw_set_last_error(FSW_OK);
 }
 
+FSW_STATUS fsw_stop_monitor(const FSW_HANDLE handle)
+{
+  try
+  {
+    FSW_SESSION *session = get_session(handle);
+
+    if (session->monitor == nullptr)
+      return fsw_set_last_error(int(FSW_ERR_UNKNOWN_MONITOR_TYPE));
+
+    if (!session->monitor->is_running())
+      return fsw_set_last_error(int(FSW_OK));
+
+    session->monitor->stop();
+  }
+  catch (libfsw_exception& ex)
+  {
+    return fsw_set_last_error(int(ex));
+  }
+  catch (int error)
+  {
+    return fsw_set_last_error(error);
+  }
+
+  return fsw_set_last_error(FSW_OK);
+}
+
 FSW_STATUS fsw_destroy_session(const FSW_HANDLE handle)
 {
   int ret = FSW_OK;
 
   try
   {
-    SESSION_GUARD;
-
     FSW_SESSION *session = get_session(handle);
 
-#ifdef HAVE_CXX_MUTEX
-#  ifdef HAVE_CXX_UNIQUE_PTR
-    const unique_ptr<mutex>& sm = session_mutexes[handle];
-    lock_guard<mutex> sm_lock(*sm.get());
-#  else
-    mutex * sm = session_mutexes[handle];
-    lock_guard<mutex> sm_lock(*sm);
-#  endif
-#endif
-
     if (session->monitor)
     {
+      if (session->monitor->is_running())
+      {
+        return fsw_set_last_error(FSW_ERR_MONITOR_ALREADY_RUNNING);
+      }
+
       void *context = session->monitor->get_context();
 
       if (!context)
@@ -929,30 +787,19 @@ FSW_STATUS fsw_destroy_session(const FSW_HANDLE handle)
       delete session->monitor;
     }
 
-    sessions.erase(handle);
+    delete session;
   }
   catch (int error)
   {
     ret = error;
   }
 
-#ifdef HAVE_CXX_MUTEX
-  session_mutexes.erase(handle);
-#endif
-
   return fsw_set_last_error(ret);
 }
 
 FSW_SESSION *get_session(const FSW_HANDLE handle)
 {
-  if (sessions.find(handle) == sessions.end())
-    throw int(FSW_ERR_SESSION_UNKNOWN);
-
-#ifdef HAVE_CXX_UNIQUE_PTR
-  return sessions[handle].get();
-#else
-  return sessions[handle];
-#endif
+  return handle;
 }
 
 FSW_STATUS fsw_set_last_error(const FSW_STATUS error)
